import typing as t
from functools import lru_cache

import numpy as np
import plotly.graph_objs as go
from Bio import Phylo

from cellarium.cas.postprocessing.ontology_aware import (
    CAS_CL_LABEL_PROPERTY_REF,
    CAS_FRACTION_PROPERTY_REF,
    CAS_SCORE_PROPERTY_REF,
)

DEFAULT_SCORE_COLORSCALE = [
    [0.0, "rgb(214, 47, 38)"],
    [0.1, "rgb(214, 47, 38)"],
    [0.2, "rgb(244, 109, 67)"],
    [0.3, "rgb(252, 172, 96)"],
    [0.4, "rgb(254, 224, 139)"],
    [0.5, "rgb(254, 254, 189)"],
    [0.6, "rgb(217, 239, 139)"],
    [0.7, "rgb(164, 216, 105)"],
    [0.8, "rgb(102, 189, 99)"],
    [0.9, "rgb(63, 170, 89)"],
    [1.0, "rgb(25, 151, 79)"],
]


class CircularTreePlot:
    def __init__(
        self,
        tree: Phylo.BaseTree.Tree,
        title: t.Optional[str] = None,
        order: str = "preorder",
        dist: float = 1.0,
        start_angle: float = 0.0,
        end_angle: float = 360.0,
        start_leaf: str = "first",
        min_node_radius: float = 6.0,
        max_node_radius: float = 16.0,
        linecolor: str = "rgb(20,20,20)",
        score_colorscale: list = DEFAULT_SCORE_COLORSCALE,
        shown_cl_names_set: set[str] = set(),
    ):
        """
        Initialize the CircularTreePlot class.

        :param tree: An instance of Bio.Phylo.Newick.Tree or Bio.Phylo.PhyloXML.Phylogeny.
        :param order: Tree traversal method to associate polar coordinates to its nodes.
        :param dist: Vertical distance between two consecutive leaves in the rectangular tree layout.
        :param start_angle: Angle in degrees representing the angle of the first leaf mapped to a circle.
        :param end_angle: Angle in degrees representing the angle of the last leaf.
        :param start_leaf: Keyword with two possible values:
            'first' - to map the leaves in the list tree.get_terminals() onto a circle in the counter-clockwise direction.
            'last' - to map the leaves in the reversed list tree.get_terminals()[::-1].
        :param min_node_radius: Minimum radius of a node (the limit of 0. fraction of cells having that node in CAS response)
        :param max_node_radius: Minimum radius of a node (the limit of 1. fraction of cells having that node in CAS response)
        :param score_colorscale: Colorscale for scores
        """
        self.tree = tree
        self.title = title
        self.order = order
        self.dist = dist
        self.start_angle = np.deg2rad(start_angle)
        self.end_angle = np.deg2rad(end_angle)
        self.start_leaf = start_leaf
        self.min_node_radius = min_node_radius
        self.max_node_radius = max_node_radius
        self.linecolor = linecolor
        self.score_colorscale = score_colorscale
        self.shown_cl_names_set = shown_cl_names_set

        # generate circular tree coordinate data
        (
            self.x_nodes,
            self.y_nodes,
            self.x_lines,
            self.y_lines,
            self.x_arcs,
            self.y_arcs,
        ) = self._get_circular_tree_data()

        # generate assets for visualization
        self.tooltip_string_list = []
        self.score_list = []
        self.cl_name_list = []
        self.cl_label_list = []
        self.size_list = []
        self.clade_index_to_cl_name_map = dict()
        self.clade_to_index = dict()
        self.cl_name_to_cl_label_map = dict()
        for idx, clade in enumerate(self.tree.find_clades(order="preorder")):
            cl_name = clade.name
            self.clade_index_to_cl_name_map[idx] = cl_name
            self.clade_to_index[clade] = idx
            parsed_property_dict = dict()
            for property in clade.properties:
                if property.ref == CAS_SCORE_PROPERTY_REF:
                    parsed_property_dict["score"] = float(property.value)
                elif property.ref == CAS_FRACTION_PROPERTY_REF:
                    parsed_property_dict["fraction"] = float(property.value)
                elif property.ref == CAS_CL_LABEL_PROPERTY_REF:
                    parsed_property_dict["cl_label"] = property.value
                else:
                    raise ValueError
            assert "score" in parsed_property_dict
            assert "fraction" in parsed_property_dict
            assert "cl_label" in parsed_property_dict
            cl_label = parsed_property_dict["cl_label"]
            score = parsed_property_dict["score"]
            fraction = parsed_property_dict["fraction"]
            tooltip_string = f"ID: {cl_name}<br>Label: {cl_label}<br>Evidence score: {score:.5f}<br>Cell fraction: {fraction * 100:.3f}%"
            size = self.min_node_radius + (self.max_node_radius - self.min_node_radius) * fraction
            self.tooltip_string_list.append(tooltip_string)
            self.score_list.append(score)
            self.size_list.append(size)
            self.cl_name_list.append(cl_name)
            self.cl_label_list.append(cl_label)
            self.cl_name_to_cl_label_map[cl_name] = cl_label

        self.num_nodes = len(self.clade_to_index.keys())
        # text labels
        self.text_labels = [
            self.cl_name_to_cl_label_map[cl_name] if cl_name in shown_cl_names_set else ""
            for cl_name in self.cl_name_list
        ]

    def _get_radius(self) -> t.Dict:
        """
        Associates each clade root with its radius, equal to the distance from that clade to the tree root.

        :return: Dictionary {clade: node_radius}
        """
        node_radius = self.tree.depths()

        if not np.count_nonzero(list(node_radius.values())):
            node_radius = self.tree.depths(unit_branch_lengths=True)
        return node_radius

    def _get_vertical_position(self) -> t.Dict:
        """
        Returns a dictionary {clade: y_coord}, where y_coord is the Cartesian y-coordinate of a clade root in a rectangular phylogram.

        :return: Dictionary {clade: y_coord}
        """
        if self.start_leaf == "first":
            node_ycoord = {leaf: k for k, leaf in enumerate(self.tree.get_terminals())}
        elif self.start_leaf == "last":
            node_ycoord = {leaf: k for k, leaf in enumerate(reversed(self.tree.get_terminals()))}
        else:
            raise ValueError("start_leaf can be only 'first' or 'last'")

        def assign_ycoord(clade):
            for subclade in clade:
                if subclade not in node_ycoord:
                    assign_ycoord(subclade)
            node_ycoord[clade] = 0.5 * (node_ycoord[clade.clades[0]] + node_ycoord[clade.clades[-1]])

        if self.tree.root.clades:
            assign_ycoord(self.tree.root)
        return node_ycoord

    def _ycoord_to_theta(self, y: float, ymin: float, ymax: float) -> float:
        """
        Maps a y-coordinate to an angle in radians.

        :param y: y-coordinate.
        :param ymin: Minimum y-coordinate.
        :param ymax: Maximum y-coordinate.
        :return: Corresponding angle in radians.
        """
        return self.start_angle + (self.end_angle - self.start_angle) * (y - ymin) / float(ymax - ymin)

    def _get_points_on_lines(
        self,
        line_type: str,
        x_left: float = 0,
        x_right: float = 0,
        y_right: float = 0,
        y_bottom: float = 0,
        y_top: float = 0,
    ) -> t.Tuple[t.List[float], t.List[float]]:
        """
        Define the points that generate a radial branch and the circular arcs, perpendicular to that branch.

        :param line_type: Type of line ('radial' or 'angular').
        :param x_left: Left x-coordinate.
        :param x_right: Right x-coordinate.
        :param y_right: Right y-coordinate.
        :param y_bottom: Bottom y-coordinate.
        :param y_top: Top y-coordinate.
        :return: Tuple containing lists of x and y coordinates of the line representative points.
        """
        if line_type == "radial":
            theta = self._ycoord_to_theta(y_right, self.ymin, self.ymax)
            X = [x_left * np.cos(theta), x_right * np.cos(theta), None]
            Y = [x_left * np.sin(theta), x_right * np.sin(theta), None]
        elif line_type == "angular":
            theta_bottom = self._ycoord_to_theta(y_bottom, self.ymin, self.ymax)
            theta_top = self._ycoord_to_theta(y_top, self.ymin, self.ymax)
            t = np.linspace(0, 1, 10)
            theta = (1 - t) * theta_bottom + t * theta_top
            X = list(x_right * np.cos(theta)) + [None]
            Y = list(x_right * np.sin(theta)) + [None]
        else:
            raise ValueError("line_type can be only 'radial' or 'angular'")

        return X, Y

    def _get_line_lists(
        self,
        clade: Phylo.BaseTree.Clade,
        x_left: float,
        x_lines: t.List[float],
        y_lines: t.List[float],
        x_arcs: t.List[float],
        y_arcs: t.List[float],
    ):
        """
        Recursively compute the lists of points that span the tree branches.

        :param clade: Clade of the tree.
        :param x_left: Left x-coordinate.
        :param x_lines: List of x-coordinates of radial edge ends.
        :param y_lines: List of y-coordinates of radial edge ends.
        :param x_arcs: List of x-coordinates of arc segments for tree branches.
        :param y_arcs: List of y-coordinates of arc segments for tree branches.
        """
        x_right = self.node_radius[clade]
        y_right = self.node_ycoord[clade]

        X, Y = self._get_points_on_lines(line_type="radial", x_left=x_left, x_right=x_right, y_right=y_right)
        x_lines.extend(X)
        y_lines.extend(Y)

        if clade.clades:
            y_top = self.node_ycoord[clade.clades[0]]
            y_bottom = self.node_ycoord[clade.clades[-1]]

            X, Y = self._get_points_on_lines(line_type="angular", x_right=x_right, y_bottom=y_bottom, y_top=y_top)
            x_arcs.extend(X)
            y_arcs.extend(Y)

            for child in clade:
                self._get_line_lists(child, x_right, x_lines, y_lines, x_arcs, y_arcs)

    def _get_circular_tree_data(
        self,
    ) -> t.Tuple[t.List[float], t.List[float], t.List[float], t.List[float], t.List[float], t.List[float]]:
        """
        Define data needed to get the Plotly plot of a circular tree.

        :return: Tuple containing xnodes, ynodes, xlines, ylines, xarc, yarc.
        """
        self.node_radius = self._get_radius()
        self.node_ycoord = self._get_vertical_position()
        y_vals = self.node_ycoord.values()
        self.ymin, self.ymax = min(y_vals), max(y_vals)
        self.ymin -= self.dist

        x_lines = []
        y_lines = []
        x_arcs = []
        y_arcs = []
        self._get_line_lists(self.tree.root, 0, x_lines, y_lines, x_arcs, y_arcs)

        x_nodes = []
        y_nodes = []

        for clade in self.tree.find_clades(order=self.order):
            theta = self._ycoord_to_theta(self.node_ycoord[clade], self.ymin, self.ymax)
            x_nodes.append(self.node_radius[clade] * np.cos(theta))
            y_nodes.append(self.node_radius[clade] * np.sin(theta))

        return x_nodes, y_nodes, x_lines, y_lines, x_arcs, y_arcs

    @property
    @lru_cache(maxsize=None)
    def plotly_figure(self) -> go.Figure:
        # Plot points
        trace_nodes = dict(
            type="scatter",
            name="data_nodes",
            x=self.x_nodes,
            y=self.y_nodes,
            mode="markers+text",
            text=self.text_labels,
            textposition="top center",
            textfont=dict(size=10),  # Base font size
            marker=dict(
                color=self.score_list,
                size=self.size_list,
                colorscale=self.score_colorscale,
                colorbar=dict(
                    thickness=10,
                    dtick=10,
                    ticklen=4,
                    title=dict(
                        text="Evidence score",
                        side="top",
                        font=dict(family="Arial", size=14),
                    ),
                    tickvals=[0, 0.25, 0.5, 0.75, 1],
                    ticktext=["0", "0.25", "0.5", "0.75", "1"],
                    tickmode="array",
                    lenmode="fraction",
                    len=0.4,
                    orientation="h",  # Horizontal orientation
                    xanchor="right",
                    yanchor="bottom",
                    x=1.0,
                    y=-0.1,
                    tickfont=dict(family="Arial", size=12),
                ),
                cmin=0.0,
                cmax=1.0,
                opacity=1.0,
            ),
            hovertext=self.tooltip_string_list,
            hoverinfo="text",
            opacity=1,
            # uirevision is needed to maintain pan/zoom state.  It must be updated to trigger a refresh
            uirevision="true",
        )

        # Straight lines
        trace_radial_lines = dict(
            type="scatter",
            name="radial_lines",
            x=self.x_lines,
            y=self.y_lines,
            mode="lines",
            line=dict(color=self.linecolor, width=1),
            hoverinfo="none",
        )

        # Curved lines
        trace_arcs = dict(
            type="scatter",
            name="arc_lines",
            x=self.x_arcs,
            y=self.y_arcs,
            mode="lines",
            line=dict(color=self.linecolor, width=1, shape="spline"),
            hoverinfo="none",
        )

        title = dict(
            text=self.title,
            font=dict(family="Arial", size=1) if self.title is None else dict(family="Arial", size=20),
        )
        layout = dict(
            title=title,
            font=dict(family="Arial", size=14),
            autosize=True,
            showlegend=False,
            xaxis=dict(
                visible=False,
                scaleanchor="y",
                scaleratio=1,
            ),
            yaxis=dict(visible=False),
            hovermode="closest",
            dragmode="pan",
            plot_bgcolor="rgb(255,255,255)",
            margin=dict(l=0, r=25, t=50, b=0),
        )

        fig = go.Figure(data=[trace_radial_lines, trace_arcs, trace_nodes], layout=layout)

        return fig

    def get_clade_path_from_index(self, selected_cl_idx: int) -> t.Optional[t.List[str]]:
        """
        Returns the path from the root to the selected clade.  Returns None if the index isn't found.
        """
        node = list(self.tree.find_clades(order="preorder"))[selected_cl_idx]
        if node is None:
            return None

        return [self.tree.root.name] + [cl.name for cl in self.tree.root.get_path(target=node)]

    def update_selected_nodes(self, selected_cl_path: t.List[str] = []):
        if len(selected_cl_path) == 0:
            line_widths = 0
        else:
            # Find the starting node to start navigating down the tree
            node = self.tree.find_clades(name=selected_cl_path[0], order="preorder")
            selected_nodes = []
            if node is not None:
                node = next(node)
                selected_nodes.append(self.clade_to_index[node])

            for node_name in selected_cl_path[1:]:
                node = next(filter(lambda cl: cl.name == node_name, node.clades))
                if node is not None:
                    selected_nodes.append(self.clade_to_index[node])
                else:
                    break
            last_index = selected_nodes.pop()

            line_widths = [0] * self.num_nodes
            for idx in selected_nodes:
                line_widths[idx] = 1.5
            # The selected note should have a thicker line
            line_widths[last_index] = 3

        self.plotly_figure.update_traces(
            marker=dict(line=dict(color="DarkSlateGrey", width=line_widths)),
            selector=({"name": "data_nodes"}),
            uirevision=self.plotly_figure["layout"]["uirevision"],
        )



import typing as t
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from enum import Enum
from operator import itemgetter

import numpy as np
import scipy.sparse as sp
from anndata import AnnData

from .cell_ontology.cell_ontology_cache import CL_CELL_ROOT_NODE, CL_EUKARYOTIC_CELL_ROOT_NODE, CellOntologyCache
from .common import get_obs_indices_for_cluster

# AnnData-related constants
CAS_CL_SCORES_ANNDATA_OBSM_KEY = "cas_cl_scores"
CAS_METADATA_ANNDATA_UNS_KEY = "cas_metadata"

# PhyloXML-related constants
CAS_SCORE_PROPERTY_REF = "CAS:score"
CAS_FRACTION_PROPERTY_REF = "CAS:fraction"
CAS_CL_LABEL_PROPERTY_REF = "CAS:cl_label"


def convert_cas_ontology_aware_response_to_score_matrix(
    adata: AnnData, cas_ontology_aware_response: list, cl: CellOntologyCache
) -> sp.csr_matrix:
    """
    Generate a sparse matrix of CAS ontology-aware scores.

    This function takes an AnnData object, a list of CAS ontology-aware responses, and a CellOntologyCache object
    and generates a sparse matrix of CAS ontology-aware scores. The sparse matrix represents the evidence scores
    of all ontology terms (columns) for each cell (rows).

    :param adata: An AnnData object containing the query cells.
    :type adata: AnnData

    :param cas_ontology_aware_response: A list of CAS ontology-aware responses.
    :type cas_ontology_aware_response: list

    :param cl: A CellOntologyCache object containing the cell ontology information.
    :type cl: CellOntologyCache

    :return: A sparse matrix of CAS ontology-aware scores.
    :rtype: sp.csr_matrix
    """
    row = []
    col = []
    data = []

    obs_values = adata.obs.index.values
    for obs_idx, cas_cell_response in enumerate(cas_ontology_aware_response):
        assert cas_cell_response["query_cell_id"] == obs_values[obs_idx]
        for match in cas_cell_response["matches"]:
            row.append(obs_idx)
            col.append(cl.cl_names_to_idx_map[match["cell_type_ontology_term_id"]])
            data.append(match["score"])

    n_obs = len(cas_ontology_aware_response)
    n_cl_names = len(cl.cl_names)
    return sp.coo_matrix((data, (row, col)), shape=(n_obs, n_cl_names)).tocsr()


def insert_cas_ontology_aware_response_into_adata(
    cas_ontology_aware_response: list, adata: AnnData, cl: CellOntologyCache
) -> None:
    """
    Inserts Cellarium CAS ontology aware response into `obsm` property of a provided AnnData file as a
    :class:`scipy.sparse.csr_matrix` named `cas_cl_scores`.

    :param cas_ontology_aware_response: The Cellarium CAS ontology aware response.
    :type cas_ontology_aware_response: list

    :param adata: The AnnData object to insert the response into.
    :type adata: AnnData

    :param cl: The CellOntologyCache object containing cell ontology term names and labels.
    :type cl: CellOntologyCache

    :return: None
    """
    adata.obsm[CAS_CL_SCORES_ANNDATA_OBSM_KEY] = convert_cas_ontology_aware_response_to_score_matrix(
        adata, cas_ontology_aware_response, cl
    )
    adata.uns[CAS_METADATA_ANNDATA_UNS_KEY] = {"cl_names": cl.cl_names, "cl_labels": cl.cl_labels}


class CellOntologyScoresAggregationOp(Enum):
    """The aggregation operation to apply to the scores of each ontology term."""

    MEAN = 1
    MEDIAN = 2


class CellOntologyScoresAggregationDomain(Enum):
    """The domain over which to aggregate the scores of each ontology term."""

    ALL_CELLS = 1
    OVER_THRESHOLD = 2


@dataclass
class AggregatedCellOntologyScores:
    """The aggregated ontology scores and the fraction of cells that have non-zero scores for each term."""

    aggregation_op: CellOntologyScoresAggregationOp
    aggregation_domain: CellOntologyScoresAggregationDomain
    threshold: float
    aggregated_scores_c: np.ndarray
    fraction_over_threshold_c: np.ndarray
    cl_labels: list
    cl_names: list


def get_aggregated_cas_ontology_aware_scores(
    adata: AnnData,
    obs_indices: t.Optional[t.Sequence],
    aggregation_op: CellOntologyScoresAggregationOp = CellOntologyScoresAggregationOp.MEAN,
    aggregation_domain: CellOntologyScoresAggregationDomain = CellOntologyScoresAggregationDomain.ALL_CELLS,
    threshold: float = 1e-4,
) -> AggregatedCellOntologyScores:
    """
    Aggregates Cellarium CAS ontology-aware scores over a provided group of cells.

    :param adata: Annotated data object containing CAS scores.
    :param obs_indices: Indices of the cells to consider for aggregation. If None, all cells are considered.
    :param aggregation_op: The aggregation operation to apply to the CAS scores.
      Default is CellOntologyScoresAggregationOp.MEAN.
    :param aggregation_domain: The domain over which to perform the aggregation.
      Default is CellOntologyScoresAggregationDomain.ALL_CELLS.
    :param threshold: The threshold value for considering a CAS score as non-zero.
      Default is 1e-4.
    :return: An instance of AggregatedCellOntologyScores.
    """
    assert CAS_CL_SCORES_ANNDATA_OBSM_KEY in adata.obsm

    if obs_indices is not None:
        assert len(obs_indices) > 0
        sliced_cas_scores_nc = adata.obsm[CAS_CL_SCORES_ANNDATA_OBSM_KEY][obs_indices]
        n_selected_cells = len(obs_indices)
    else:  # all cells
        sliced_cas_scores_nc = adata.obsm[CAS_CL_SCORES_ANNDATA_OBSM_KEY]
        n_selected_cells = adata.n_obs

    mask_aggregation_nc = (sliced_cas_scores_nc > threshold).toarray()
    fraction_over_threshold_c = mask_aggregation_nc.sum(0) / n_selected_cells

    if aggregation_op == CellOntologyScoresAggregationOp.MEAN:
        aggregation_op_func = np.mean
    elif aggregation_op == CellOntologyScoresAggregationOp.MEDIAN:
        aggregation_op_func = np.median
    else:
        raise ValueError

    sliced_cas_scores_dense_nc = sliced_cas_scores_nc.toarray()
    n_cols = sliced_cas_scores_dense_nc.shape[1]
    aggregated_scores_c = np.zeros((n_cols,))
    for c in range(n_cols):
        if aggregation_domain == CellOntologyScoresAggregationDomain.ALL_CELLS:
            value = aggregation_op_func(sliced_cas_scores_dense_nc[:, c])
        elif aggregation_domain == CellOntologyScoresAggregationDomain.OVER_THRESHOLD:
            data = sliced_cas_scores_dense_nc[:, c][mask_aggregation_nc[:, c]]
            value = aggregation_op_func(data) if data.size else 0.0
        else:
            raise ValueError
        aggregated_scores_c[c] = value

    return AggregatedCellOntologyScores(
        aggregation_op=aggregation_op,
        aggregation_domain=aggregation_domain,
        threshold=threshold,
        aggregated_scores_c=aggregated_scores_c,
        fraction_over_threshold_c=fraction_over_threshold_c,
        cl_labels=adata.uns[CAS_METADATA_ANNDATA_UNS_KEY]["cl_labels"],
        cl_names=adata.uns[CAS_METADATA_ANNDATA_UNS_KEY]["cl_names"],
    )


def convert_aggregated_cell_ontology_scores_to_rooted_tree(
    aggregated_scores: AggregatedCellOntologyScores,
    cl: CellOntologyCache,
    root_cl_name: str = CL_EUKARYOTIC_CELL_ROOT_NODE,
    min_fraction: float = 0.0,
    hidden_cl_names_set: t.Optional[t.Set[str]] = None,
) -> OrderedDict:
    """
    Convert aggregated cell ontology scores to a rooted tree.

    :param aggregated_scores: The aggregated cell ontology scores.
    :type aggregated_scores: AggregatedCellOntologyScores
    :param cl: The cell ontology cache.
    :type cl: CellOntologyCache
    :param root_cl_name: The name of the root cell ontology.
    :type root_cl_name: str
    :param min_fraction: The minimum fraction threshold, defaults to 0.0.
    :type min_fraction: float, optional
    :param hidden_cl_names_set: A set of hidden cell ontology names, defaults to an empty set.
    :type hidden_cl_names_set: set[str], optional
    :return: The rooted tree structure as an ordered dictionary.
    :rtype: OrderedDict
    """

    if hidden_cl_names_set is None:
        hidden_cl_names_set = set()

    tree_dict = OrderedDict()
    score_dict = {
        cl_name: score for cl_name, score in zip(aggregated_scores.cl_names, aggregated_scores.aggregated_scores_c)
    }
    fraction_dict = {
        cl_name: score
        for cl_name, score in zip(aggregated_scores.cl_names, aggregated_scores.fraction_over_threshold_c)
    }

    def build_subtree(node_dict: OrderedDict, node_name: str) -> OrderedDict:
        if node_name in hidden_cl_names_set:
            return node_dict
        if score_dict[node_name] <= aggregated_scores.threshold:
            return node_dict
        if fraction_dict[node_name] <= min_fraction:
            return node_dict
        node_dict[node_name] = {"score": score_dict[node_name], "fraction": fraction_dict[node_name]}
        children_nodes = list(cl.cl_graph.successors(node_name))
        if len(children_nodes) > 0:
            node_dict[node_name]["children"] = OrderedDict()
            for children_node_name in children_nodes:
                build_subtree(node_dict[node_name]["children"], children_node_name)
        return node_dict

    tree_dict = build_subtree(tree_dict, root_cl_name)
    # Validate that this is actually a rooted tree and if not recalculate with the base cell node
    if len(tree_dict) == 1:  # singly-rooted tree
        return tree_dict
    elif root_cl_name != CL_CELL_ROOT_NODE:
        return convert_aggregated_cell_ontology_scores_to_rooted_tree(
            aggregated_scores=aggregated_scores,
            cl=cl,
            root_cl_name=CL_CELL_ROOT_NODE,
            min_fraction=min_fraction,
            hidden_cl_names_set=hidden_cl_names_set,
        )
    else:
        raise ValueError("The tree is not singly-rooted.")


def generate_phyloxml_from_scored_cell_ontology_tree(
    tree_dict: OrderedDict, tree_name: str, cl: CellOntologyCache, indent: int = 3
) -> str:
    """
    Generate a PhyloXML string representation of a scored cell ontology tree.

    :param tree_dict: The dictionary representation of the scored cell ontology tree.
    :type tree_dict: OrderedDict
    :param tree_name: The name of the tree.
    :type tree_name: str
    :param cl: The CellOntologyCache object containing the cell ontology information.
    :type cl: CellOntologyCache
    :param indent: The number of spaces to use for indentation. Defaults to 3.
    :type indent: int, optional
    :return: The PhyloXML string representation of the scored cell ontology tree.
    :rtype: str
    """

    assert indent > 0
    assert len(tree_dict) == 1  # singly-rooted tree
    root_node_name = list(tree_dict.keys())[0]

    def ind(level: int):
        return " " * indent * level

    phyloxml_header = (
        '<?xml version="1.0" encoding="UTF-8"?>\n'
        '<phyloxml xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.phyloxml.org http://www.phyloxml.org/1.00/phyloxml.xsd" xmlns="http://www.phyloxml.org">\n'
        '<phylogeny rooted="true">\n'
        f"{ind(1)}<name>{tree_name}</name>\n"
    )

    phyloxml_footer = "</phylogeny>\n</phyloxml>\n"

    def _get_subtree_phyloxml_string(subtree_dict: OrderedDict, node_name: str, level: int) -> str:
        phyloxml_string = f"{ind(level)}<clade>\n" f"{ind(level + 1)}<name>{node_name}</name>\n"
        if level > 1:
            phyloxml_string += f"{ind(level + 1)}<branch_length>1.0</branch_length>\n"
        phyloxml_string += (
            f'{ind(level + 1)}<property datatype="xsd:string" ref="{CAS_CL_LABEL_PROPERTY_REF}" applies_to="clade">{cl.cl_names_to_labels_map[node_name]}</property>\n'
            f'{ind(level + 1)}<property datatype="xsd:float" ref="{CAS_SCORE_PROPERTY_REF}" applies_to="clade">{subtree_dict[node_name]["score"]}</property>\n'
            f'{ind(level + 1)}<property datatype="xsd:float" ref="{CAS_FRACTION_PROPERTY_REF}" applies_to="clade">{subtree_dict[node_name]["fraction"]}</property>\n'
        )
        if "children" in subtree_dict[node_name]:
            for child_node_name in subtree_dict[node_name]["children"].keys():
                phyloxml_string += _get_subtree_phyloxml_string(
                    subtree_dict[node_name]["children"], child_node_name, level + 1
                )
        phyloxml_string += f"{ind(level)}</clade>\n"
        return phyloxml_string

    phyloxml_string = phyloxml_header + _get_subtree_phyloxml_string(tree_dict, root_node_name, 1) + phyloxml_footer
    return phyloxml_string


def get_most_granular_top_k_calls(
    aggregated_scores: AggregatedCellOntologyScores, cl: CellOntologyCache, min_acceptable_score: float, top_k: int = 1
) -> t.List[tuple]:
    depth_list = list(
        map(cl.get_longest_path_lengths_from_target(CL_EUKARYOTIC_CELL_ROOT_NODE).get, aggregated_scores.cl_names)
    )
    sorted_score_and_depth_list = sorted(
        list(
            (score, depth, cl_name)
            for score, depth, cl_name in zip(
                aggregated_scores.aggregated_scores_c, depth_list, aggregated_scores.cl_names
            )
            if score >= min_acceptable_score
        ),
        key=itemgetter(1),
        reverse=True,
    )
    trunc_list = sorted_score_and_depth_list[:top_k]
    # pad with root node if necessary
    for _ in range(len(trunc_list) - top_k):
        trunc_list.append((1.0, 0, CL_EUKARYOTIC_CELL_ROOT_NODE))
    return trunc_list


def compute_most_granular_top_k_calls_single(
    adata: AnnData,
    cl: CellOntologyCache,
    min_acceptable_score: float,
    top_k: int = 3,
    obs_prefix: str = "cas_cell_type",
):
    top_k_calls_dict = defaultdict(list)
    scores_array_nc = adata.obsm[CAS_CL_SCORES_ANNDATA_OBSM_KEY].toarray()

    # make a wrapped aggregated scores dataclass for the single cell in question
    aggregated_scores = AggregatedCellOntologyScores(
        aggregation_op=CellOntologyScoresAggregationOp.MEAN,
        aggregation_domain=CellOntologyScoresAggregationDomain.ALL_CELLS,
        threshold=0.0,
        aggregated_scores_c=np.zeros(
            (
                len(
                    cl.cl_names,
                )
            )
        ),
        fraction_over_threshold_c=np.ones(
            (
                len(
                    cl.cl_names,
                )
            )
        ),
        cl_labels=cl.cl_labels,
        cl_names=cl.cl_names,
    )

    for i_cell in range(adata.n_obs):
        aggregated_scores.aggregated_scores_c = scores_array_nc[i_cell]
        top_k_output = get_most_granular_top_k_calls(aggregated_scores, cl, min_acceptable_score, top_k)
        for k in range(top_k):
            top_k_calls_dict[f"{obs_prefix}_score_{k + 1}"].append(top_k_output[k][0])
            top_k_calls_dict[f"{obs_prefix}_name_{k + 1}"].append(top_k_output[k][2])
            top_k_calls_dict[f"{obs_prefix}_label_{k + 1}"].append(cl.cl_names_to_labels_map[top_k_output[k][2]])

    for k, v in top_k_calls_dict.items():
        adata.obs[k] = v


def compute_most_granular_top_k_calls_cluster(
    adata: AnnData,
    cl: CellOntologyCache,
    min_acceptable_score: float,
    cluster_label_obs_column: str,
    aggregation_op: CellOntologyScoresAggregationOp = CellOntologyScoresAggregationOp.MEAN,
    aggregation_domain: CellOntologyScoresAggregationDomain = CellOntologyScoresAggregationDomain.ALL_CELLS,
    aggregation_score_threshod: float = 1e-4,
    top_k: int = 3,
    obs_prefix: str = "cas_cell_type",
):
    top_k_calls_dict = dict()
    for k in range(top_k):
        top_k_calls_dict[f"{obs_prefix}_score_{k + 1}"] = [None] * adata.n_obs
        top_k_calls_dict[f"{obs_prefix}_name_{k + 1}"] = [None] * adata.n_obs
        top_k_calls_dict[f"{obs_prefix}_label_{k + 1}"] = [None] * adata.n_obs

    def _update_list(target_list, indices, value):
        for idx in indices:
            target_list[idx] = value

    for cluster_label in adata.obs[cluster_label_obs_column].cat.categories:
        obs_indices = get_obs_indices_for_cluster(adata, cluster_label_obs_column, cluster_label)
        aggregated_scores = get_aggregated_cas_ontology_aware_scores(
            adata, obs_indices, aggregation_op, aggregation_domain, aggregation_score_threshod
        )
        top_k_output = get_most_granular_top_k_calls(aggregated_scores, cl, min_acceptable_score, top_k)
        for k in range(top_k):
            _update_list(top_k_calls_dict[f"{obs_prefix}_score_{k + 1}"], obs_indices, top_k_output[k][0])
            _update_list(top_k_calls_dict[f"{obs_prefix}_name_{k + 1}"], obs_indices, top_k_output[k][2])
            _update_list(
                top_k_calls_dict[f"{obs_prefix}_label_{k + 1}"],
                obs_indices,
                cl.cl_names_to_labels_map[top_k_output[k][2]],
            )

    for k, v in top_k_calls_dict.items():
        adata.obs[k] = v



from cellarium.cas.client import CASClient

from . import constants, exceptions, postprocessing, preprocessing, service, settings, version, visualization

__version__ = version.get_version()

__all__ = [
    "CASClient",
    "preprocessing",
    "postprocessing",
    "visualization",
    "settings",
    "constants",
    "exceptions",
    "service",
    "settings",
    "version",
]



from dataclasses import dataclass


class CASBaseError(Exception):
    """Base class for all CAS exceptions"""


class CASClientError(CASBaseError):
    pass


class HTTPError(CASBaseError):
    pass


class HTTPError5XX(HTTPError):
    pass


class HTTPError404(HTTPError):
    pass


class HTTPError403(HTTPError):
    pass


class HTTPError401(HTTPError):
    pass


class HTTPClientError(HTTPError):
    pass


@dataclass
class DataValidationError(CASBaseError):
    missing_features: int
    extra_features: int


class QuotaExceededError(CASBaseError):
    pass


class ClientTooOldError(CASBaseError):
    pass



_API_BASE_ENDPOINT = "api"
_API_CELL_OPERATIONS_ENDPOINT_PREFIX = f"{_API_BASE_ENDPOINT}/cellarium-cell-operations"
_API_GENERAL_ENDPOINT_PREFIX = f"{_API_BASE_ENDPOINT}/cellarium-general"
# General
VALIDATE_TOKEN = f"{_API_GENERAL_ENDPOINT_PREFIX}/validate-token"
APPLICATION_INFO = f"{_API_GENERAL_ENDPOINT_PREFIX}/application-info"
FEEDBACK_OPT_OUT = f"{_API_GENERAL_ENDPOINT_PREFIX}/feedback/opt-out"
FEEDBACK_ANSWER = f"{_API_GENERAL_ENDPOINT_PREFIX}/feedback/answer?client_session_id={{client_session_id}}&client_action_id={{client_action_id}}"
GET_FEATURE_SCHEMAS = f"{_API_GENERAL_ENDPOINT_PREFIX}/feature-schemas"
GET_SCHEMA_BY_NAME = f"{_API_GENERAL_ENDPOINT_PREFIX}/feature-schema/{{schema_name}}"
LIST_MODELS = f"{_API_GENERAL_ENDPOINT_PREFIX}/list-models"
GET_USER_QUOTA = f"{_API_GENERAL_ENDPOINT_PREFIX}/quota"
VALIDATE_VERSION = f"{_API_GENERAL_ENDPOINT_PREFIX}/validate-client-version"
# Cell Analysis
ANNOTATE_CELL_TYPE_SUMMARY_STATS_STRATEGY = (
    f"{_API_CELL_OPERATIONS_ENDPOINT_PREFIX}/annotate-cell-type-summary-statistics-strategy"
)
ANNOTATE_CELL_TYPE_ONTOLOGY_AWARE_STRATEGY = (
    f"{_API_CELL_OPERATIONS_ENDPOINT_PREFIX}/annotate-cell-type-ontology-aware-strategy"
)
NEAREST_NEIGHBOR_SEARCH = f"{_API_CELL_OPERATIONS_ENDPOINT_PREFIX}/nearest-neighbor-search"
QUERY_CELLS_BY_IDS = f"{_API_CELL_OPERATIONS_ENDPOINT_PREFIX}/query-cells-by-ids"



import logging

try:
    from .circular_tree_plot_umap_dash_app.app import CASCircularTreePlotUMAPDashApp  # noqa
    from .ui_utils import find_and_kill_process  # noqa

    __all__ = ["CASCircularTreePlotUMAPDashApp", "find_and_kill_process"]
except ImportError:
    logging.debug(
        """
Visualization dependencies not installed.
To install the Cellarium CAS Client with visualation dependencies, please run:
pip install --force-reinstall cellarium-cas[vis]
"""
    )
    __all__ = []



import typing as t

import anndata
import numpy as np

from cellarium.cas import constants

TOTAL_MRNA_UMIS_COLUMN_NAME = "total_mrna_umis"


def calculate_total_mrna_umis(adata: anndata.AnnData, count_matrix_input: constants.CountMatrixInput) -> None:
    """
    Calculate the total mRNA UMIs (Unique Molecular Identifiers) for each observation in the AnnData object and add them
    as a new column in the ``.obs`` attribute. It is recommended to use this callback before data sanitization to
    calculate all expressed genes.

    :param adata: The annotated data matrix from which to calculate the total mRNA UMIs.
    :param count_matrix_input: Where to obtain a feature expression count matrix from. Choice of: 'X', 'raw.X'

    :return: A new AnnData object with the same data as ``adata`` but with an additional field in ``.obs`` containing
        the total mRNA UMIs for each observation.

    .. note::
       The function assumes that the ``.X`` attribute of the input AnnData is a count matrix where each element i,j
       represents the count of unique transcripts (UMIs) for the ith observation and jth variable (gene). The function
       performs an inline operation on the input AnnData object, changing it in place.

    Example:
    ________
        >>> calculate_total_mrna_umis(adata=adata)
        >>> 'total_mrna_umis' in adata.obs.columns
        True.
    """
    count_matrix = adata.X if count_matrix_input == constants.CountMatrixInput.X else adata.raw.X
    adata.obs[TOTAL_MRNA_UMIS_COLUMN_NAME] = np.array(count_matrix.sum(axis=1)).flatten()


_PRE_SANITIZE_CALLBACKS: t.List[t.Callable[[anndata.AnnData, constants.CountMatrixInput], None]] = [
    calculate_total_mrna_umis,
]


def pre_sanitize_callback(adata: anndata.AnnData, count_matrix_input: constants.CountMatrixInput) -> None:
    """
    Apply each necessary callback before data sanitization

    :param adata: Input :class:`anndata.AnnData` instance
    :param count_matrix_input: Where to obtain a feature expression count matrix from. Choice of: 'X', 'raw.X'

    :return: A new :class:`anndata.AnnData` instance after callback modifications
    """
    for callback in _PRE_SANITIZE_CALLBACKS:
        callback(adata, count_matrix_input)
